(function(window) {
	var Piece = function(canvas, config)
	{
		this.initialize(canvas, config);
	}
	var p = Piece.prototype = new BasePiece();
	//
	p.initialize = function(canvas, config)
	{
		BasePiece.prototype.initialize.call(this, canvas, config);
		this.initSound();
	}

	p.onKeyUp = function(e)
	{
		BasePiece.prototype.onKeyUp.call(this, e);
		if (!this.config.debug) return;
		var c = String.fromCharCode(e.which);
		if (c=="R") this.reset();
	}
		
	
	/*********************************
	 *		    INTERACTION
	 ********************************/
	
	p.initInteraction = function()
	{
		this.pane.addEventListener("mousedown", this.handleMouseDown.bind(this));
		this.pane.addEventListener("pressmove", this.handleMouseMove.bind(this));
		this.pane.addEventListener("pressup", this.handleMouseUp.bind(this));
	}
	p.handleMouseDown = function(e)
	{
		if (this.pointerID) return;
		this.pointerID = e.pointerID;
		const x = e.stageX, y = e.stageY;
		this.dragStart = this.dragLast = this.tickLast;
		this.dragPos = new createjs.Point(x,y);
	}
	p.handleMouseMove = function(e)
	{
		if (e.pointerID!=this.pointerID) return;
		const x = e.stageX, y = e.stageY, now = this.tickLast;
		const cfg = this.config, w = this.width, h = this.height, cfgp = cfg.pane;
		let d = cfg.orientation == Config.ORIENTATION.X ? (x - this.dragPos.x) / w : (y - this.dragPos.y) / h;
		this.dragPos.setValues(x,y);
		this.dragLast = now;
		//Check dock:
		if (this.lastOffset!=0)
		{
			//check if we cross 0 offset
			if (Math.abs(Math.sgn(this.offset + d) != Math.sgn(this.lastOffset)))
			{				
				if (this.dockStart == null) 
				{
					//not yet docked: dock
					this.dockStart = now;//track time of docking
					d = -this.offset;//dock at 0
					this.stopSoundSlide();
				} else if (now - this.dockStart < this.config.dock.duration) {
					//dock period not elapsed
					d = 0;
				} else { 
					//dock period has elapsed
					this.dockStart = null;
					this.lastOffset = this.offset + d;
				}
			} else {
				this.dockStart = null;
				this.lastOffset = this.offset + d;
			}
		}
		else if (now - this.dragStart > this.config.dock.delay)
		{
			//track offset after 'grace period' to accomodate for tap inaccuracy
			this.lastOffset = this.offset + d;
		}
		if (d==0) return;
		//Stop sound if we just changed direction
		if (this.lastDelta != 0 && Math.sgn(this.lastDelta) != Math.sgn(d)) {
			this.stopSoundSlide();
		}
		this.lastDelta = d;
		//
		var min = (cfgp.moveMode&1)==1 ? -cfgp.sizeOpening : 0;
		var max = (cfgp.moveMode&2)==2 ? cfgp.sizeOpening : 0;
		var offset = Math.min(max, Math.max(min, this.offset + d));
		if (offset == this.offset) //no change
		{			
			this.stopSoundSlide();
		}
		else
		{
			this.offset = offset;
			this.drawPane();
			this.updateVolume();
			if (!this.soundSlide && this.dockStart == null) {
				this.playSoundSlide();
			}
		}
	}
	p.handleMouseUp = function(e)
	{
		if (e.pointerID!=this.pointerID) return;
		this.pointerID = null;
		this.stopDrag();
	}
	p.stopDrag = function() {
		this.lastOffset = this.offset;
		this.dockStart = null;
		this.dragLast = null;
		this.dragStart = null;
		this.stopSoundSlide();
	}
	

	 
	/*********************************
	 *			    FLOW
	 ********************************/
	
	p.setSize = function(w,h,dpr)
	{
		this.dpr = dpr;
		w = this.width = Math.floor(w*dpr);
		h = this.height = Math.floor(h*dpr);
		//
		log("setSize",this.width, this.height, this.dpr);
		this.reset();
	}

	p.start = function()
	{
		BasePiece.prototype.start.apply(this);
		this.sky = this.stage.addChild(new createjs.Shape());
		this.pane = this.stage.addChild(new createjs.Shape());
		this.pane.alpha = this.config.pane.alpha;
		this.initInteraction();
	}
	
	p.reset = function()
	{
		const cfg = this.config, w = this.width, h = this.height;
		//	
		this.stopDrag();
		this.offset = this.lastOffset = cfg.pane.offsetInitial;	
		//
		const g = this.sky.graphics.c();
		createLinearGradientFromStops(g, cfg.sky.gradient.stops,0,0,0,h);
		g.r(0,0,w,h);
		//
		this.drawPane();
	}
	
	p.update = function()
	{
		if (this.dragLast != null && this.tickLast - this.dragLast> this.config.sound.slideStopAfterInactive) {
			this.stopSoundSlide();
		}		
		return true;
	}	
		
	p.drawPane = function() {
		var cfg = this.config, w = this.width, h = this.height;
		const g = this.pane.graphics.c();
		const x = cfg.pane.gradientMoveFactor * this.offset * w;
		createLinearGradientFromStops(g, cfg.pane.gradient.stops,x,0,x+w,h);
		let px = 0, py = 0;
		if (cfg.orientation == Config.ORIENTATION.X) {
			px = this.offset * w;
		} else {
			py = this.offset * h;
		}
		g.r(px,py,w,h);
	}

	p.checkDock = function(d) {
		//check if we have to dock the pane at offset 0
		const now = this.tickLast;
		if (this.lastOffset!=0)
		{
			//check if we cross 0 offset
			if (Math.abs(Math.sgn(this.offset + d) - Math.sgn(this.lastOffset)) == 2)
			{				
				if (this.dockStart == null) //not yet docked: dock
				{
					log(1, now);
					this.dockStart = now;//track time of docking
					d = -this.offset;//dock at 0
					this.stopSoundSlide();
				} else if (now - this.dockStart < this.config.dock.duration) {
					d = 0;//dock period not elapsed: dont do anything
				} else { //dock period has elapsed: reset dock and track new offset
					log(2, now);
					this.dockStart = null;//reset dockStart
					this.lastOffset = this.offset + d;
				}
			}
		}
		else if (now - this.dragStart > this.config.dock.delay)
		{
			//track offset after 'grace period' to accomodate for tap inaccuracy
			this.lastOffset = this.offset + d;
		}
		return d;
	}


	/*********************************
	 *		    SOUND
	 ********************************/

	p.initSound = function() {
		const sounds = [this.config.sound.birds, this.config.sound.slide];
		SoundManager.init(sounds, this.onSoundsLoaded.bind(this));
	}
	p.onSoundsLoaded = function() {
		log("onSoundsLoaded");
		this.handleClickToInitiateSound = this.handleClickToInitiateSound.bind(this);
		this.stage.addEventListener("stagemousedown", this.handleClickToInitiateSound);
	}
	p.handleClickToInitiateSound = function() {
		log("handleClickToInitiateSound");
		this.soundAmbient = this.playSound(this.config.sound.birds, {loop:-1});
		this.updateVolume();
		this.stage.removeEventListener("stagemousedown", this.handleClickToInitiateSound);
	}

	p.playSound = function(id, props) {
		if (!this.config.sound.isOn) return;
		return SoundManager.play(id, props);
	}

	p.playSoundSlide = function() {
		const cfgs = this.config.sound;
		this.soundSlide = this.playSound(cfgs.slide, {volume:cfgs.slideVolume});
		this.soundSlide.addEventListener("complete", function() {
			this.soundSlide = null;
		}.bind(this));
	}
	p.stopSoundSlide = function() {
		if (this.soundSlide != null) {
			SoundManager.stopSound(this.soundSlide);
			this.soundSlide = null;
		}
	}
	p.updateVolume = function() {
		if (!this.soundAmbient) return;
		//calc volume based on pane offset
		const cfg = this.config;
		const r = Math.abs(this.offset) / cfg.pane.sizeOpening;
		const vol = r * (1-cfg.sound.volumeMin) + cfg.sound.volumeMin;
		this.soundAmbient.volume = vol;
	}


	///

	const createLinearGradientFromStops = function(g, stops, x0,y0,x1,y1) {
		const cs = [], rs = [];
		stops.forEach(stop => { 
			cs.push(stop[0]);
			rs.push(stop[1]);
		});
		g.lf(cs, rs, x0,y0, x1,y1);
	}

	window.Piece = Piece;	

}(window));

